@@ -23,6 +23,7 @@
    Author: Raph Levien <raph@artofcode.com>
 */
 #include <string.h>
+#include <math.h>
 
 #include "rsvg-shapes.h"
 #include "rsvg-styles.h"
@@ -30,6 +31,7 @@
 #include "rsvg-private.h"
 #include "rsvg-bpath-util.h"
 #include "rsvg-path.h"
+#include "rsvg-defs.h"
 
 #include <libart_lgpl/art_affine.h>
 #include <libart_lgpl/art_vpath_bpath.h>
@@ -43,6 +45,14 @@
 /* 4/3 * (1-cos 45)/sin 45 = 4/3 * sqrt(2) - 1 */
 #define RSVG_ARC_MAGIC ((double) 0.5522847498)
 
+struct _RsvgDefsPath {
+ 	RsvgDefVal super;
+	RsvgState  state;
+ 	char       *d;
+};
+ 
+typedef struct _RsvgDefsPath RsvgDefsPath;
+
 /**
  * rsvg_close_vpath: Close a vector path.
  * @src: Source vector path.
@@ -105,6 +115,43 @@ rsvg_close_vpath (const ArtVpath *src)
 	return result;
 }
 
+/* calculates how big an svp is */
+static ArtIRect
+rsvg_calculate_svp_bounds (const ArtSVP *svp)
+{
+	int i, j;	
+	int bigx, littlex, bigy, littley, assignedonce;
+	ArtIRect output;
+
+	bigx = littlex = bigy = littley = assignedonce = 0;	
+
+	for (i = 0; i < svp->n_segs; i++)
+		for (j = 0; j < svp->segs[i].n_points; j++)
+			{
+				if (!assignedonce)
+					{
+						bigx = svp->segs[i].points[j].x;
+						littlex = svp->segs[i].points[j].x;
+						bigy = svp->segs[i].points[j].y; 
+						littley = svp->segs[i].points[j].y;
+						assignedonce = 1;
+					}
+				if (svp->segs[i].points[j].x > bigx)
+					bigx = svp->segs[i].points[j].x;
+				if (svp->segs[i].points[j].x < littlex)
+					littlex = svp->segs[i].points[j].x;
+				if (svp->segs[i].points[j].y > bigy)
+					bigy = svp->segs[i].points[j].y; 
+				if (svp->segs[i].points[j].y < littley)
+					littley = svp->segs[i].points[j].y;
+			}
+	output.x0 = littlex;
+	output.y0 = littley;
+	output.x1 = bigx;
+	output.y1 = bigy;
+	return output;
+}
+
 /**
  * rsvg_render_svp: Render an SVP.
  * @ctx: Context in which to render.
@@ -121,7 +168,11 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
 	GdkPixbuf *pixbuf;
 	ArtRender *render;
 	gboolean has_alpha;
-	
+	ArtIRect temprect;
+	RsvgPSCtx gradctx;
+	RsvgState *state;
+	int i;	
+
 	pixbuf = ctx->pixbuf;
 	if (pixbuf == NULL)
 		{
@@ -144,7 +195,19 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
 	
 	art_render_svp (render, svp);
 	art_render_mask_solid (render, (opacity << 8) + opacity + (opacity >> 7));
-	rsvg_render_paint_server (render, ps, NULL); /* todo: paint server ctx */
+
+	temprect = rsvg_calculate_svp_bounds(svp);
+
+	gradctx.x0 = temprect.x0;
+	gradctx.y0 = temprect.y0;
+	gradctx.x1 = temprect.x1;
+	gradctx.y1 = temprect.y1;
+
+	state = rsvg_state_current(ctx);
+	for (i = 0; i < 6; i++)
+		gradctx.affine[i] = state->affine[i];
+
+	rsvg_render_paint_server (render, ps, &gradctx);
 	art_render_invoke (render);
 }
 
@@ -167,7 +230,7 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
 			return;
 		}
 	
-	state = &ctx->state[ctx->n_state - 1];
+	state = rsvg_state_current (ctx);
 
 	/* todo: handle visibility stuff earlier for performance benefits 
 	 * handles all path based shapes. will handle text and images separately
@@ -197,7 +260,11 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
 			svp = art_svp_from_vpath (closed_vpath);
 			g_free (closed_vpath);
 			
-			swr = art_svp_writer_rewind_new (ART_WIND_RULE_NONZERO);
+			if (state->fill_rule == FILL_RULE_EVENODD)
+				swr = art_svp_writer_rewind_new (ART_WIND_RULE_ODDEVEN);
+			else /* state->fill_rule == FILL_RULE_NONZERO */
+				swr = art_svp_writer_rewind_new (ART_WIND_RULE_NONZERO);
+
 			art_svp_intersector (svp, swr);
 			
 			svp2 = art_svp_writer_rewind_reap (swr);
@@ -261,6 +328,38 @@ rsvg_render_path(RsvgHandle *ctx, const char *d)
 	rsvg_bpath_def_free (bpath_def);
 }
 
+static void 
+rsvg_defs_path_free (RsvgDefVal *self)
+{
+	RsvgDefsPath *z = (RsvgDefsPath *)self;
+	rsvg_state_finalize (&z->state);
+	g_free (z);
+}
+
+/**
+ * Todo: Possibly publicly export this (for text)
+ */
+static void
+rsvg_handle_path (RsvgHandle *ctx, const char * d, const char * id)
+{
+	if (!ctx->in_defs)
+		rsvg_render_path (ctx, d);
+	else {
+		RsvgDefsPath *path;
+
+		if (id == NULL) 
+			return;
+
+		path = g_new (RsvgDefsPath, 1);
+		path->d = g_strdup(d);
+		rsvg_state_clone (&path->state, rsvg_state_current (ctx));
+		path->super.type = RSVG_DEF_PATH;
+		path->super.free = rsvg_defs_path_free;
+		
+		rsvg_defs_set (ctx->defs, id, &path->super);
+	}
+}
+
 void
 rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 {
@@ -268,9 +367,6 @@ rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 	char *d = NULL;
 	const char * klazz = NULL, * id = NULL;
 	
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
@@ -287,8 +383,8 @@ rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
 	if (d == NULL)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "path", klazz, id, atts);
-	rsvg_render_path (ctx, d);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "path", klazz, id, atts);
+	rsvg_handle_path (ctx, d, id);
 }
 
 static GString *
@@ -328,9 +424,6 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 	gchar ** pointlist = NULL;
 	const char * klazz = NULL, * id = NULL;
 
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-	
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
@@ -348,7 +441,7 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 	if (!verts)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
 	
 	/* todo: make the following more memory and CPU friendly */
 	g = rsvg_make_poly_point_list (verts);
@@ -367,9 +460,9 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 			if (!is_polyline)
 				g_string_append (d, "Z");
 			
-			rsvg_render_path (ctx, d->str);
-			g_string_free (d, TRUE);
 			g_strfreev(pointlist);
+			rsvg_handle_path (ctx, d->str, id);
+			g_string_free (d, TRUE);
 		}
 }
 
@@ -392,33 +485,33 @@ rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
 	double x1 = 0, y1 = 0, x2 = 0, y2 = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
+	double font_size;
 
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-
-	state = &ctx->state[ctx->n_state - 1];
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "x1"))
-						x1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y1"))
-						y1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					if (!strcmp ((char *)atts[i], "x2"))
-						x2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y2"))
-						y2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
 						id = (const char *)atts[i + 1];
 				}      
 		}
-	rsvg_parse_style_attrs (ctx, "line", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "line", klazz, id, atts);
 	
 	/* emulate a line using a path */
 	/* ("M %f %f L %f %f", x1, y1, x2, y2) */
@@ -432,7 +525,7 @@ rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
 	g_string_append_c (d, ' ');
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), y2));    
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -440,36 +533,36 @@ void
 rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 {
 	int i;
-	double x = 0., y = 0., w = -1, h = -1, rx = 0., ry = 0.;
+	double x = 0., y = 0., w = 0, h = 0, rx = 0., ry = 0.;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	gboolean got_rx = FALSE, got_ry = FALSE;
+	double font_size;
 
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-
-	state = &ctx->state[ctx->n_state - 1];
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "x"))
-						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "y"))
-						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "width"))
-						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "height"))
-						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "rx")) {
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 						got_rx = TRUE;
 					}
 					else if (!strcmp ((char *)atts[i], "ry")) {
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 						got_ry = TRUE;
 					}
 					else if (!strcmp ((char *)atts[i], "class"))
@@ -484,15 +577,15 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 	else if (got_ry && !got_rx)
 		rx = ry;	
 
-	if (x < 0. || y < 0. || w <= 0. || h <= 0. || rx < 0. || ry < 0.)
+	if (w == 0. || h == 0. || rx < 0. || ry < 0.)
 		return;
 
-	if (rx > (w / 2.))
-		rx = w / 2.;
-	if (ry > (h / 2.))
-		ry = h / 2.;
+	if (rx > fabs(w / 2.))
+		rx = fabs(w / 2.);
+	if (ry > fabs(h / 2.))
+		ry = fabs(h / 2.);
 	
-	rsvg_parse_style_attrs (ctx, "rect", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "rect", klazz, id, atts);
 	
 	/* incrementing y by 1 properly draws borders. this is a HACK */
 	y += 1.;
@@ -576,7 +669,7 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 	g_string_append_c (d, ' ');
 	g_string_append (d, g_ascii_dtostr (buf, sizeof (buf), y));
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -587,26 +680,26 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 	double cx = 0, cy = 0, r = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
-
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-
-	state = &ctx->state[ctx->n_state - 1];
+	double font_size;
+	
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "r"))
 						r = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, 
 															  rsvg_viewport_percentage((gdouble)ctx->width, (gdouble)ctx->height), 
-															  state->font_size);
+															  font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
@@ -614,10 +707,10 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 				}
 		}
 	
-	if (cx < 0. || cy < 0. || r <= 0.)
+	if (r <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "circle", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "circle", klazz, id, atts);
 	
 	/* approximate a circle using 4 bezier curves */
 
@@ -680,7 +773,7 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 
 	g_string_append (d, " Z");
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -691,26 +784,26 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 	double cx = 0, cy = 0, rx = 0, ry = 0;
 	GString * d = NULL;
 	const char * klazz = NULL, * id = NULL;
-	RsvgState *state;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
-
-	/* skip over defs entries for now */
-	if (ctx->in_defs) return;
-
-	state = &ctx->state[ctx->n_state - 1];
+	double font_size;
+	
+	if (ctx->n_state > 0)
+		font_size = rsvg_state_current (ctx)->font_size;
+	else
+		font_size = 12.0;
 
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "rx"))
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
 					else if (!strcmp ((char *)atts[i], "ry"))
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
 					else if (!strcmp ((char *)atts[i], "class"))
 						klazz = (const char *)atts[i + 1];
 					else if (!strcmp ((char *)atts[i], "id"))
@@ -718,10 +811,10 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 				}
 		}
 	
-	if (cx < 0. || cy < 0. || rx <= 0. || ry <= 0.)
+	if (rx <= 0. || ry <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "ellipse", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "ellipse", klazz, id, atts);
 	
 	/* approximate an ellipse using 4 bezier curves */
 
@@ -784,7 +877,7 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
 
 	g_string_append (d, " Z");
 
-	rsvg_render_path (ctx, d->str);
+	rsvg_handle_path (ctx, d->str, id);
 	g_string_free (d, TRUE);
 }
 
@@ -812,7 +905,7 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	/* skip over defs entries for now */
 	if (ctx->in_defs) return;
 
-	state = &ctx->state[ctx->n_state - 1];
+	state = rsvg_state_current (ctx);
 	
 	if (atts != NULL)
 		{
@@ -839,7 +932,7 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	if (!href || x < 0. || y < 0. || w <= 0. || h <= 0.)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, "image", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, state, "image", klazz, id, atts);
 	
 	/* figure out if image is visible or not */
 	if (!state->visible)
@@ -904,3 +997,73 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 	g_object_unref (G_OBJECT (img));
 	g_free (rgb);
 }
+
+void 
+rsvg_start_use (RsvgHandle *ctx, const xmlChar **atts)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	const char * klazz = NULL, *id = NULL, *xlink_href = NULL;
+	double x = 0, y = 0, width = 0, height = 0;
+	int i;
+	double affine[6];
+	gboolean got_width = FALSE, got_height = FALSE;
+
+	if (atts != NULL)
+		{
+			for (i = 0; atts[i] != NULL; i += 2)
+				{
+					if (!strcmp ((char *)atts[i], "x"))
+						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+					else if (!strcmp ((char *)atts[i], "y"))
+						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+					else if (!strcmp ((char *)atts[i], "width")) {
+						width = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						got_width = TRUE;
+					}
+					else if (!strcmp ((char *)atts[i], "height")) {
+						height = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+						got_height = TRUE;
+					}
+					else if (!strcmp ((char *)atts[i], "class"))
+						klazz = (const char *)atts[i + 1];
+					else if (!strcmp ((char *)atts[i], "id"))
+						id = (const char *)atts[i + 1];
+					else if (!strcmp ((char *)atts[i], "xlink:href"))
+						xlink_href = (const char *)atts[i + 1];
+				}
+		}
+	
+	/* < 0 is an error, 0 disables rendering. TODO: handle positive values correctly */
+	if (got_width || got_height)
+		if (width <= 0. || height <= 0.)
+			return;
+	
+	if (xlink_href != NULL)
+		{
+			RsvgDefVal * parent = rsvg_defs_lookup (ctx->defs, xlink_href+1);
+			if (parent != NULL)
+				switch(parent->type)
+					{
+					case RSVG_DEF_PATH:
+						{
+							RsvgDefsPath *path = (RsvgDefsPath*)parent;
+
+							/* combine state definitions */
+							rsvg_state_clone (state, &path->state);
+							art_affine_translate (affine, x, y);
+							art_affine_multiply (state->affine, affine, state->affine);
+
+							rsvg_parse_style_attrs (ctx, state, "use", klazz, id, atts);
+							if (state->opacity != 0xff)
+								rsvg_push_opacity_group (ctx);
+
+							/* always want to render inside of a <use/> */
+							rsvg_render_path (ctx, path->d);
+							break;
+						}
+					default:
+						g_warning ("Unhandled defs entry/type %s %d\n", id, parent->type);
+						return;
+					}
+		}
+}
