@@ -45,6 +45,17 @@ enum {
 	TEXT_STRIKE    = 0x04
 };
 
+typedef enum {
+	TEXT_ANCHOR_START,
+	TEXT_ANCHOR_MIDDLE,
+	TEXT_ANCHOR_END
+} TextAnchor;
+
+enum {
+	FILL_RULE_EVENODD = 0,
+	FILL_RULE_NONZERO = 1
+};
+
 typedef struct {
 	double affine[6];
 	
@@ -53,6 +64,8 @@ typedef struct {
 	RsvgPaintServer *fill;
 	gint fill_opacity; /* 0..255 */
 	
+	gint fill_rule;	
+
 	RsvgPaintServer *stroke;
 	gint stroke_opacity; /* 0..255 */
 	double stroke_width;
@@ -62,13 +75,16 @@ typedef struct {
 	ArtPathStrokeJoinType join;
 	
 	double         font_size;
-	char        *  font_family;
+	char         * font_family;
+	char         * lang;
 	PangoStyle     font_style;
 	PangoVariant   font_variant;
 	PangoWeight    font_weight;
 	PangoStretch   font_stretch;
 	TextDecoration font_decor;
-	
+	PangoDirection text_dir;
+	TextAnchor     text_anchor;	
+
 	guint text_offset;
 	
 	guint32 stop_color; /* rgb */
@@ -90,16 +106,18 @@ void rsvg_parse_style_pair (RsvgHandle *ctx, RsvgState *state,
 							const char *key, const char *val);
 void rsvg_parse_style (RsvgHandle *ctx, RsvgState *state, const char *str);
 void rsvg_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen);
-void rsvg_parse_style_attrs (RsvgHandle *ctx, const char * tag,
+
+void rsvg_parse_style_attrs (RsvgHandle *ctx, RsvgState *state, const char * tag,
 							 const char * klazz, const char * id,
 							 const xmlChar **atts);
 
 gdouble rsvg_viewport_percentage (gdouble width, gdouble height);
-
 void rsvg_pop_opacity_group (RsvgHandle *ctx, int opacity);
 void rsvg_push_opacity_group (RsvgHandle *ctx);
 gboolean rsvg_parse_transform (double dst[6], const char *src);
 
+RsvgState * rsvg_state_current (RsvgHandle *ctx);
+
 G_END_DECLS
 
 #endif /* RSVG_STYLES_H */
