@@ -32,6 +32,8 @@
 #include <libart_lgpl/art_affine.h>
 #include <libart_lgpl/art_render_mask.h>
 
+#include <pango/pangoft2.h>
+
 #include "rsvg-shapes.h"
 
 #if 0
@@ -121,8 +123,10 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	PangoLayout *layout;
 	PangoFontDescription *font;
 	PangoLayoutLine *line;
-	PangoRectangle ink_rect, line_ink_rect;
+	PangoRectangle ink_rect, line_ink_rect, logical_rect;
 	FT_Bitmap bitmap;
+	RsvgPSCtx gradctx;
+	int i;
 	
 	state = rsvg_state_current (ctx);
 	if (state->fill == NULL && state->font_size <= 0)
@@ -169,8 +173,16 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 		}
 	
 	if (ctx->pango_context == NULL)
-		ctx->pango_context = pango_ft2_get_context ((guint)ctx->dpi, (guint)ctx->dpi);   
-		
+		{
+			PangoFT2FontMap *fontmap;
+
+			fontmap = PANGO_FT2_FONT_MAP (pango_ft2_font_map_new ());
+			pango_ft2_font_map_set_resolution (fontmap, ctx->dpi, ctx->dpi);
+
+			ctx->pango_context = pango_ft2_font_map_create_context (fontmap);
+			g_object_unref (fontmap);
+		}
+
 	layout = pango_layout_new (ctx->pango_context);
 	pango_layout_set_text (layout, string, -1);
 	font = pango_font_description_copy (pango_context_get_font_description (ctx->pango_context));
@@ -209,8 +221,34 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	bitmap.pixel_mode = ft_pixel_mode_grays;
 	
 	pango_ft2_render_layout (&bitmap, layout, -ink_rect.x, -ink_rect.y);
+	pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
+	if ((state->text_dir == PANGO_DIRECTION_RTL) ||
+		(state->text_dir == PANGO_DIRECTION_LTR)) {
+		switch (state->text_anchor) {
+		case TEXT_ANCHOR_MIDDLE:
+			logical_rect.width /= 2;
+			break;
+		case TEXT_ANCHOR_END:
+			break;
+		default:
+			logical_rect.width = 0;
+			break;
+		}
+		logical_rect.height = 0;
+	} else {
+		switch (state->text_anchor) {
+		case TEXT_ANCHOR_MIDDLE:
+			logical_rect.height /= 2;
+			break;
+		case TEXT_ANCHOR_END:
+			break;
+		default:
+			logical_rect.height = 0;
+			break;
+		}
+		logical_rect.width = 0;
+	}
 	g_object_unref (layout);
-
 	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
 	render = art_render_new (0, 0,
 							 gdk_pixbuf_get_width (pixbuf),
@@ -223,16 +261,24 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 							 has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
 							 NULL);
 	
-	rsvg_render_paint_server (render, state->fill, NULL); /* todo: paint server ctx */
+	gradctx.x0 = line_ink_rect.x;
+	gradctx.y0 = line_ink_rect.y;
+	gradctx.x1 = line_ink_rect.x + logical_rect.width;
+	gradctx.y1 = line_ink_rect.y + logical_rect.height;
+	for (i = 0; i < 6; i++)
+		gradctx.affine[i] = state->affine[i];
+	
+	rsvg_render_paint_server (render, state->fill, &gradctx);
+
 	opacity = state->fill_opacity * state->opacity;
 	opacity = opacity + (opacity >> 7) + (opacity >> 14);
 	
 	art_render_mask_solid (render, opacity);
 	art_render_mask (render,
-					 state->affine[4] + line_ink_rect.x + state->text_offset,
-					 state->affine[5] + line_ink_rect.y,
-					 state->affine[4] + line_ink_rect.x + bitmap.width + state->text_offset,
-					 state->affine[5] + line_ink_rect.y + bitmap.rows,
+					 state->affine[4] + line_ink_rect.x + state->text_offset - logical_rect.width,
+					 state->affine[5] + line_ink_rect.y - logical_rect.height,
+					 state->affine[4] + line_ink_rect.x + bitmap.width + state->text_offset - logical_rect.width,
+					 state->affine[5] + line_ink_rect.y + bitmap.rows - logical_rect.height,
 					 bitmap.buffer, bitmap.pitch);
 	art_render_invoke (render);
 	
