@@ -82,7 +82,8 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 	gint new_width, new_height;
 	double x_zoom = 1.;
 	double y_zoom = 1.;
-	
+	double affine[6];
+
 	double vbox_x = 0, vbox_y = 0, vbox_w = 0, vbox_h = 0;
 	gboolean has_vbox = FALSE;
 
@@ -94,11 +95,11 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 					   at least for now, but i'll include them here anyway */
 					if (!strcmp ((char *)atts[i], "width"))
 						width = rsvg_css_parse_length ((char *)atts[i + 1], ctx->dpi, &percent, &em, &ex);
-					else if (!strcmp ((char *)atts[i], "height"))	    
+					else if (!strcmp ((char *)atts[i], "height"))
 						height = rsvg_css_parse_length ((char *)atts[i + 1], ctx->dpi, &percent, &em, &ex);
-					else if (!strcmp ((char *)atts[i], "x"))	    
+					else if (!strcmp ((char *)atts[i], "x"))
 						x = rsvg_css_parse_length ((char *)atts[i + 1], ctx->dpi, &percent, &em, &ex);
-					else if (!strcmp ((char *)atts[i], "y"))	    
+					else if (!strcmp ((char *)atts[i], "y"))
 						y = rsvg_css_parse_length ((char *)atts[i + 1], ctx->dpi, &percent, &em, &ex);
 					else if (!strcmp ((char *)atts[i], "viewBox"))
 						{
@@ -106,12 +107,12 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 															&vbox_w, &vbox_h);
 						}
 				}
-			
+
 			if (has_vbox && vbox_w > 0. && vbox_h > 0.)
 				{
 					new_width  = (int)floor (vbox_w);
 					new_height = (int)floor (vbox_h);
-					
+
 					/* apply the sizing function on the *original* width and height
 					   to acquire our real destination size. we'll scale it against
 					   the viewBox's coordinates later */
@@ -122,17 +123,17 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 				{
 					new_width  = width;
 					new_height = height;
-					
+
 					/* apply the sizing function to acquire our new width and height.
 					   we'll scale this against the old values later */
 					if (ctx->size_func)
 						(* ctx->size_func) (&new_width, &new_height, ctx->user_data);
 				}
-			
+
 			/* set these here because % are relative to viewbox */
 			ctx->width = new_width;
 			ctx->height = new_height;
-			
+
 			if (!has_vbox)
 				{
 					  x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
@@ -140,39 +141,35 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 				}
 			else
 				{
-#if 1
 					x_zoom = (width < 0 || new_width < 0) ? 1 : (double) width / new_width;
 					y_zoom = (height < 0 || new_height < 0) ? 1 : (double) height / new_height;
-#else
-					x_zoom = (width < 0 || new_width < 0) ? 1 : (double) new_width / width;
-					y_zoom = (height < 0 || new_height < 0) ? 1 : (double) new_height / height;	  
-#endif
-					
+
 					/* reset these so that we get a properly sized SVG and not a huge one */
 					new_width  = (width == -1 ? new_width : width);
 					new_height = (height == -1 ? new_height : height);
 				}
 
 			/* Scale size of target pixbuf */
-			state = &ctx->state[ctx->n_state - 1];
-			art_affine_scale (state->affine, x_zoom, y_zoom);
-			
-#if 0
-			if (vbox_x != 0. || vbox_y != 0.)
+			state = rsvg_state_current (ctx);
+
+			art_affine_identity (state->affine);
+
+			if (has_vbox && (vbox_x != 0. || vbox_y != 0.))
 				{
-					double affine[6];
-					art_affine_translate (affine, vbox_x, vbox_y);
-					art_affine_multiply (state->affine, affine, state->affine);
+					art_affine_translate (affine, - vbox_x, - vbox_y);
+					art_affine_multiply (state->affine, state->affine, affine);
 				}
-#endif
-			
+
+			art_affine_scale (affine, x_zoom, y_zoom);
+			art_affine_multiply (state->affine, state->affine, affine);
+
 			if (new_width < 0 || new_height < 0)
 				{
 					g_warning ("rsvg_start_svg: width and height not specified in the SVG, nor supplied by the size callback");
 					if (new_width < 0) new_width = 500;
 					if (new_height < 0) new_height = 500;
 				}
-			
+
 			if (new_width >= INT_MAX / 4)
 				{
 					/* FIXME: GError here? */
@@ -186,9 +183,9 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 					g_warning ("rsvg_start_svg: width too large");
 					return;
 				}
-			
+
 			/* FIXME: Add GError here if size is too big. */
-			
+
 			pixels = g_try_malloc (rowstride * new_height);
 			if (pixels == NULL)
 				{
@@ -210,7 +207,7 @@ rsvg_start_svg (RsvgHandle *ctx, const xmlChar **atts)
 static void
 rsvg_start_g (RsvgHandle *ctx, const xmlChar **atts)
 {
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
+	RsvgState *state = rsvg_state_current (ctx);
 	const char * klazz = NULL, * id = NULL;
 	int i;
 	
@@ -225,7 +222,7 @@ rsvg_start_g (RsvgHandle *ctx, const xmlChar **atts)
 				}
 		}
 	
-	rsvg_parse_style_attrs (ctx, "g", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, state, "g", klazz, id, atts);
 	if (state->opacity != 0xff)
 		rsvg_push_opacity_group (ctx);
 }
@@ -233,7 +230,7 @@ rsvg_start_g (RsvgHandle *ctx, const xmlChar **atts)
 static void
 rsvg_end_g (RsvgHandle *ctx)
 {
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
+	RsvgState *state = rsvg_state_current (ctx);
 	
 	if (state->opacity != 0xff)
 		rsvg_pop_opacity_group (ctx, state->opacity);
@@ -259,6 +256,10 @@ typedef struct _RsvgSaxHandlerGstops {
 	const char * parent_tag;
 } RsvgSaxHandlerGstops;
 
+/* hide this fact from the general public */
+typedef RsvgSaxHandlerDefs RsvgSaxHandlerTitle;
+typedef RsvgSaxHandlerDefs RsvgSaxHandlerDesc;
+
 static void
 rsvg_gradient_stop_handler_free (RsvgSaxHandler *self)
 {
@@ -278,10 +279,7 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 	int n_stop;
 	
 	if (strcmp ((char *)name, "stop"))
-		{
-			g_warning ("unexpected <%s> element in gradient\n", name);
-			return;
-		}
+		return;
 	
 	rsvg_state_init (&state);
 	
@@ -332,13 +330,14 @@ rsvg_gradient_stop_handler_end (RsvgSaxHandler *self, const xmlChar *name)
 {
 	RsvgSaxHandlerGstops *z = (RsvgSaxHandlerGstops *)self;
 	RsvgHandle *ctx = z->ctx;
+	RsvgSaxHandler *prev = &z->parent->super;
 	
 	if (!strcmp((char *)name, z->parent_tag))
 		{
 			if (ctx->handler != NULL)
 				{
 					ctx->handler->free (ctx->handler);
-					ctx->handler = &z->parent->super;
+					ctx->handler = prev;
 				}
 		}
 }
@@ -396,7 +395,7 @@ rsvg_linear_gradient_free (RsvgDefVal *self)
 static void
 rsvg_start_linear_gradient (RsvgHandle *ctx, const xmlChar **atts)
 {
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
+	RsvgState *state = rsvg_state_current (ctx);
 	RsvgLinearGradient *grad = NULL;
 	int i;
 	const char *id = NULL;
@@ -487,13 +486,13 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, const xmlChar **atts)
 	
 	rsvg_defs_set (ctx->defs, id, &grad->super);
 	
-	for (i = 0; i < 6; i++)
-		grad->affine[i] = state->affine[i];
-
 	if (got_transform)
-		art_affine_multiply (grad->affine, affine, grad->affine);
-	
-	/* state inherits parent/cloned information unless it's explicity gotten */
+		for (i = 0; i < 6; i++)
+			grad->affine[i] = affine[i];
+	else
+		art_affine_identity(grad->affine);
+
+	/* gradient inherits parent/cloned information unless it's explicity gotten */
 	grad->obj_bbox = (cloned && !got_bbox) ? grad->obj_bbox : obj_bbox;
 	grad->x1 = (cloned && !got_x1) ? grad->x1 : x1;
 	grad->y1 = (cloned && !got_y1) ? grad->y1 : y1;
@@ -516,7 +515,7 @@ rsvg_radial_gradient_free (RsvgDefVal *self)
 static void
 rsvg_start_radial_gradient (RsvgHandle *ctx, const xmlChar **atts, const char * tag) /* tag for conicalGradient */
 {
-	RsvgState *state = &ctx->state[ctx->n_state - 1];
+	RsvgState *state = rsvg_state_current (ctx);
 	RsvgRadialGradient *grad = NULL;
 	int i;
 	const char *id = NULL;
@@ -630,14 +629,14 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, const xmlChar **atts, const char *
 	}
 	
 	rsvg_defs_set (ctx->defs, id, &grad->super);
-	
-	for (i = 0; i < 6; i++)
-		grad->affine[i] = state->affine[i];
 
 	if (got_transform)
-		art_affine_multiply (grad->affine, affine, grad->affine);
+		for (i = 0; i < 6; i++)
+			grad->affine[i] = affine[i];
+	else
+		art_affine_identity(grad->affine);
 	
-	/* state inherits parent/cloned information unless it's explicity gotten */
+	/* gradient inherits parent/cloned information unless it's explicity gotten */
 	grad->obj_bbox = (cloned && !got_bbox) ? grad->obj_bbox : obj_bbox;
 	grad->cx = (cloned && !got_cx) ? grad->cx : cx;
 	grad->cy = (cloned && !got_cy) ? grad->cy : cy;
@@ -679,13 +678,14 @@ rsvg_style_handler_end (RsvgSaxHandler *self, const xmlChar *name)
 {
 	RsvgSaxHandlerStyle *z = (RsvgSaxHandlerStyle *)self;
 	RsvgHandle *ctx = z->ctx;
+	RsvgSaxHandler *prev = &z->parent->super;
 	
 	if (!strcmp ((char *)name, "style"))
 		{
 			if (ctx->handler != NULL)
 				{
 					ctx->handler->free (ctx->handler);
-					ctx->handler = &z->parent->super;
+					ctx->handler = prev;
 				}
 		}
 }
@@ -707,7 +707,7 @@ rsvg_start_style (RsvgHandle *ctx, const xmlChar **atts)
 	ctx->handler = &handler->super;
 }
 
-/* */
+/* start defs */
 
 static void
 rsvg_defs_handler_free (RsvgSaxHandler *self)
@@ -751,6 +751,20 @@ rsvg_defs_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 		rsvg_start_radial_gradient (ctx, atts, "conicalGradient");
 	else if (!strcmp ((char *)name, "style"))
 		rsvg_start_style (ctx, atts);
+	else if (!strcmp ((char *)name, "path"))
+		rsvg_start_path (ctx, atts);
+	else if (!strcmp ((char *)name, "line"))
+		rsvg_start_line (ctx, atts);
+	else if (!strcmp ((char *)name, "rect"))
+		rsvg_start_rect (ctx, atts);
+	else if (!strcmp ((char *)name, "circle"))
+		rsvg_start_circle (ctx, atts);
+	else if (!strcmp ((char *)name, "ellipse"))
+		rsvg_start_ellipse (ctx, atts);
+	else if (!strcmp ((char *)name, "polygon"))
+		rsvg_start_polygon (ctx, atts);
+	else if (!strcmp ((char *)name, "polyline"))
+		rsvg_start_polyline (ctx, atts);
 }
 
 static void
@@ -791,6 +805,160 @@ rsvg_start_defs (RsvgHandle *ctx, const xmlChar **atts)
 
 /* end defs */
 
+/* start desc */
+
+static void
+rsvg_desc_handler_free (RsvgSaxHandler *self)
+{
+	g_free (self);
+}
+
+static void
+rsvg_desc_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
+{
+	RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *)self;
+	RsvgHandle *ctx = z->ctx;
+
+	char * string = NULL;
+	char * utf8 = NULL;
+
+	/* This isn't quite the correct behavior - in theory, any graphics
+	   element may contain a title or desc element */
+
+	if (!ch || !len)
+		return;
+
+	string = g_strndup (ch, len);
+	if (!g_utf8_validate (string, -1, NULL))
+		{
+			utf8 = make_valid_utf8 (string);
+			g_free (string);
+			string = utf8;
+		}
+
+	g_string_append (ctx->desc, string);
+	g_free (string);
+}
+
+static void
+rsvg_desc_handler_start (RsvgSaxHandler *self, const xmlChar *name,
+						 const xmlChar **atts)
+{
+}
+
+static void
+rsvg_desc_handler_end (RsvgSaxHandler *self, const xmlChar *name)
+{
+	RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *)self;
+	RsvgHandle *ctx = z->ctx;
+	
+	if (!strcmp((char *)name, "desc"))
+		{
+			if (ctx->handler != NULL)
+				{
+					ctx->handler->free (ctx->handler);
+					ctx->handler = NULL;
+				}
+		}
+	
+	/* pop the state stack */
+	ctx->n_state--;
+	rsvg_state_finalize (&ctx->state[ctx->n_state]);
+}
+
+static void
+rsvg_start_desc (RsvgHandle *ctx, const xmlChar **atts)
+{
+	RsvgSaxHandlerDesc *handler = g_new0 (RsvgSaxHandlerDesc, 1);
+	
+	handler->super.free = rsvg_desc_handler_free;
+	handler->super.characters = rsvg_desc_handler_characters;
+	handler->super.start_element = rsvg_desc_handler_start;
+	handler->super.end_element   = rsvg_desc_handler_end;
+	handler->ctx = ctx;
+
+	ctx->handler = &handler->super;
+}
+
+/* end desc */
+
+/* start title */
+
+static void
+rsvg_title_handler_free (RsvgSaxHandler *self)
+{
+	g_free (self);
+}
+
+static void
+rsvg_title_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
+{
+	RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *)self;
+	RsvgHandle *ctx = z->ctx;
+
+	char * string = NULL;
+	char * utf8 = NULL;
+
+	/* This isn't quite the correct behavior - in theory, any graphics
+	   element may contain a title or desc element */
+
+	if (!ch || !len)
+		return;
+
+	string = g_strndup (ch, len);
+	if (!g_utf8_validate (string, -1, NULL))
+		{
+			utf8 = make_valid_utf8 (string);
+			g_free (string);
+			string = utf8;
+		}
+
+	g_string_append (ctx->title, string);
+	g_free (string);
+}
+
+static void
+rsvg_title_handler_start (RsvgSaxHandler *self, const xmlChar *name,
+						 const xmlChar **atts)
+{
+}
+
+static void
+rsvg_title_handler_end (RsvgSaxHandler *self, const xmlChar *name)
+{
+	RsvgSaxHandlerTitle *z = (RsvgSaxHandlerTitle *)self;
+	RsvgHandle *ctx = z->ctx;
+	
+	if (!strcmp((char *)name, "title"))
+		{
+			if (ctx->handler != NULL)
+				{
+					ctx->handler->free (ctx->handler);
+					ctx->handler = NULL;
+				}
+		}
+	
+	/* pop the state stack */
+	ctx->n_state--;
+	rsvg_state_finalize (&ctx->state[ctx->n_state]);
+}
+
+static void
+rsvg_start_title (RsvgHandle *ctx, const xmlChar **atts)
+{
+	RsvgSaxHandlerTitle *handler = g_new0 (RsvgSaxHandlerTitle, 1);
+	
+	handler->super.free = rsvg_title_handler_free;
+	handler->super.characters = rsvg_title_handler_characters;
+	handler->super.start_element = rsvg_title_handler_start;
+	handler->super.end_element   = rsvg_title_handler_end;
+	handler->ctx = ctx;
+
+	ctx->handler = &handler->super;
+}
+
+/* end title */
+
 static void
 rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 {
@@ -818,12 +986,10 @@ rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 				rsvg_start_svg (ctx, atts);
 			else if (!strcmp ((char *)name, "g"))
 				rsvg_start_g (ctx, atts);
+			else if (!strcmp ((char *)name, "defs"))
+				rsvg_start_defs (ctx, atts);
 			else if (!strcmp ((char *)name, "path"))
 				rsvg_start_path (ctx, atts);
-			else if (!strcmp ((char *)name, "text"))
-				rsvg_start_text (ctx, atts);
-			else if (!strcmp ((char *)name, "image"))
-				rsvg_start_image (ctx, atts);
 			else if (!strcmp ((char *)name, "line"))
 				rsvg_start_line (ctx, atts);
 			else if (!strcmp ((char *)name, "rect"))
@@ -832,12 +998,22 @@ rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 				rsvg_start_circle (ctx, atts);
 			else if (!strcmp ((char *)name, "ellipse"))
 				rsvg_start_ellipse (ctx, atts);
-			else if (!strcmp ((char *)name, "defs"))
-				rsvg_start_defs (ctx, atts);
 			else if (!strcmp ((char *)name, "polygon"))
 				rsvg_start_polygon (ctx, atts);
 			else if (!strcmp ((char *)name, "polyline"))
 				rsvg_start_polyline (ctx, atts);
+			else if (!strcmp ((char *)name, "use"))
+				rsvg_start_use (ctx, atts);
+			else if (!strcmp ((char *)name, "text"))
+				rsvg_start_text (ctx, atts);
+			else if (!strcmp ((char *)name, "image"))
+				rsvg_start_image (ctx, atts);
+			else if (!strcmp ((char *)name, "style"))
+				rsvg_start_style (ctx, atts);
+			else if (!strcmp ((char *)name, "title"))
+				rsvg_start_title (ctx, atts);
+			else if (!strcmp ((char *)name, "desc"))
+				rsvg_start_desc (ctx, atts);
 			
 			/* see conicalGradient discussion above */
 			else if (!strcmp ((char *)name, "linearGradient"))
@@ -1053,9 +1229,47 @@ rsvg_handle_free_impl (RsvgHandle *handle)
 		(* handle->user_data_destroy) (handle->user_data);
 	if (handle->pixbuf)
 		g_object_unref (handle->pixbuf);
+
+	if (handle->title)
+		g_string_free (handle->title, TRUE);
+	if (handle->desc)
+	    	g_string_free (handle->desc, TRUE);
+
 	g_free (handle);
 }
 
+/**
+ * rsvg_handle_get_title:
+ * @handle: An #RsvgHandle
+ *
+ * Returns the SVG's title in UTF-8 or %NULL. You must make a copy
+ * of this title if you wish to use it after #handle has been freed.
+ *
+ * Returns: The SVG's title
+ *
+ * Since: 2.4
+ */
+G_CONST_RETURN char *rsvg_handle_get_title (RsvgHandle *handle)
+{
+	return handle->title->str;
+}
+
+/**
+ * rsvg_handle_get_desc:
+ * @handle: An #RsvgHandle
+ *
+ * Returns the SVG's description in UTF-8 or %NULL. You must make a copy
+ * of this description if you wish to use it after #handle has been freed.
+ *
+ * Returns: The SVG's description
+ *
+ * Since: 2.4
+ */
+G_CONST_RETURN char *rsvg_handle_get_desc (RsvgHandle *handle)
+{
+	return handle->desc->str;
+}
+
 /**
  * rsvg_handle_new:
  *
@@ -1096,6 +1310,9 @@ rsvg_handle_init (RsvgHandle * handle)
 											   g_free, g_free);
 	
 	handle->ctxt = NULL;
+
+	handle->title = g_string_new (NULL);
+	handle->desc = g_string_new (NULL);
 }
 
 /**
@@ -1105,6 +1322,8 @@ rsvg_handle_init (RsvgHandle * handle)
  * Sets the DPI for the all future outgoing pixbufs. Common values are
  * 72, 90, and 300 DPI. Passing a number <= 0 to #dpi will 
  * reset the DPI to whatever the default value happens to be.
+ *
+ * Since: 2.2
  */
 void
 rsvg_set_default_dpi (double dpi)
@@ -1123,6 +1342,8 @@ rsvg_set_default_dpi (double dpi)
  * Sets the DPI for the outgoing pixbuf. Common values are
  * 72, 90, and 300 DPI. Passing a number <= 0 to #dpi will 
  * reset the DPI to whatever the default value happens to be.
+ *
+ * Since: 2.2
  */
 void
 rsvg_handle_set_dpi (RsvgHandle * handle, double dpi)
